C      PROGRAM Contour
C
C   Contours a scalar defined on a rectangular grid of points in (x,y).
C   Reads data from infile (a .grd file) and creates a .mpm file for MakePlot.
C
C   By Peter Bird, Department of Earth and Space Sciences, UCLA.
C
C   Transformed to Fortran 90 in February 2000 in order to use
C   dynamic memory allocation (and long file names).  However, the
C   program's FORTRAN77 roots show in its fixed format, upper-case
C   typing, and programming style.
C
C  (C) Copyright 1987,1992, 2000 by 
C      Peter Bird and the Regents of the University of California.
C
       CHARACTER*80 INFILE, MPMFIL
       CHARACTER*79 TITLE
       INTEGER DOWN, FILE, LFILE, PFILE, TFILE, UP
C
       REAL, DIMENSION(:,:), ALLOCATABLE :: DATA
C
       COMMON /IODEV/ FILE
       COMMON /SIZE/ SCALE, XOFSET, YOFSET
       DATA UP/3/, DOWN/2/
       DATA LFILE/11/, PFILE/11/, TFILE/11/
C      NOTE: All 3 device numbers above are the same since
C            lines, points, and text were combined in one .mpm file!
C
C      FOLLOWING PLOT AND FONT DIMENSIONS ARE IN INCHES:
       DATA SHIGH/7.50/, SWIDE/9.90/, SIDES /0.35/
       DATA HEIGHT /0.16/, RATIO /0.80/
C
       DATA HUGE /9.99E29/
C
C   INPUT SECTION
C
    1  FORMAT(A)
       WRITE(6,2)
    2  FORMAT('                             ----Contour----'
     &       //'  INPUT: Scalar floating-point data on a rectangular'
     &        ,' grid (a .grd file).'
     &        /' OUTPUT: A .mpm graphics metafile for use '
     &        ,'with -MakePlot-.'/
     &       //' This program expects a data file containing:'
     &        /'    XMin   dX   XMax    [ in 1st line ]'
     &        /'    YMin   dY   YMax    [ in 2nd line ]'
     &        /' ((Data(iry, jcx), jcx = 1, nx), iry = ny, 1, -1)'
     &        /' [ data arranged in rows (same Y);'
     &        /'   rows ordered from YMax to YMin;'
     &        /'   positions of line breaks unimportant ]'
     &        /' Enter the name of the (.grd) file to be read: ')
    5  READ(5,"(A)")INFILE
       OPEN(UNIT=10,FILE=INFILE,STATUS='OLD',IOSTAT=IOS)
       IF (IOS.NE.0) THEN
            WRITE(*,"(' ERROR: File not found (in this directory).')")
            WRITE(*,"(' Enter correct [drive:][\path\]filename: ')")
            GO TO 5
       END IF
       READ(10,*,IOSTAT=IOS)XMIN,DX,XMAX
       IF (IOS.NE.0) THEN
            WRITE(*,"(' ERROR: File does not begin with Xmin, dX, Xmax '
     &               ,'on first line.')")
            CALL PAUSE()
            STOP
       END IF
       TICSIZ=(0.05/8.)*(XMAX-XMIN)
       READ(10,*,IOSTAT=IOS)YMIN,DY,YMAX
       IF (IOS.NE.0) THEN
            WRITE(*,"(' ERROR: File does not have Ymin, dY, Ymax '
     &               ,'on second line.')")
            CALL PAUSE()
            STOP
       END IF
       NX=(XMAX-XMIN)/DX+1.5
       NY=(YMAX-YMIN)/DY+1.5
C
       ALLOCATE ( DATA (NY, NX) )
C
       READ(10,*,END=11)((DATA(IRY,JCX),JCX=1,NX),IRY=NY,1,-1)
   11  CONTINUE
       CLOSE(10)
C
       WRITE(6,12)
   12  FORMAT(' Data read.'
     &      //' Enter filename (with extension .mpm) for graphics'
     &       ,' output: '\)
       READ(5,1) MPMFIL
   14  WRITE(6,15)NX
   15 FORMAT(/' Enter two integers to limit X-range of data contoured;'
     &       /' The outer limits possible are: 1,',I3
     &       /' and your integers must differ by an even number!'
     &       /' Enter them here: '\)
       READ(5,*)NCX1,NCX2
       IF ((NCX1.LT.1).OR.(NCX1.GT.NX).OR.((NCX2-NCX1).LT.2).OR.
     &     (MOD((NCX2-NCX1),2).NE.0)) GO TO 14
   16  WRITE(6,17)NY
   17 FORMAT(/' Enter two integers to limit Y-range of data contoured;'
     &       /' The outer limits possible are: 1,',I3
     &       /' and your integers must differ by an even number!'
     &       /' Enter them here: '\)
       READ(5,*)NCY1,NCY2
       IF ((NCY1.LT.1).OR.(NCY1.GT.NY).OR.((NCY2-NCY1).LT.2).OR.
     &     (MOD((NCY2-NCY1),2).NE.0)) GO TO 16
       XBOT=XMIN+(NCX1-1)*DX
       YBOT=YMIN+(NCY1-1)*DY
       ZBOT=DATA(NCY1,NCX1)
       XTOP=XBOT
       YTOP=YBOT
       ZTOP=ZBOT
       DO 30 IRY=NCY1,NCY2
            DO 20 JCX=NCX1,NCX2
                 DATUM=DATA(IRY,JCX)
                 IF (DATUM.LT.ZBOT) THEN
                      XBOT=XMIN+(JCX-1)*DX
                      YBOT=YMIN+(IRY-1)*DY
                      ZBOT=DATUM
                 ENDIF
                 IF (DATUM.GT.ZTOP) THEN
                      XTOP=XMIN+(JCX-1)*DX
                      YTOP=YMIN+(IRY-1)*DY
                      ZTOP=DATUM
                 ENDIF
   20       CONTINUE
   30  CONTINUE
       WRITE(6,35) ZBOT,ZTOP
   35  FORMAT(/' Data ranges from ',1P,E13.5,' to ',E13.5
     &        /' Enter desired contour interval: '\)
       READ(5,*)DFCON
       OPEN(UNIT=LFILE,FILE=MPMFIL)
C
C   FIGURE SCALE & DRAW BOX AROUND REGION, WITH CROSS AT (XMIN,YMIN)
C
       X0B=XMIN+(NCX1-1)*DX
       X1B=XMIN+(NCX2-1)*DX
       Y0B=YMIN+(NCY1-1)*DY
       Y1B=YMIN+(NCY2-1)*DY
       XSCALE=(SWIDE-SIDES-SIDES)/(X1B-X0B)
       YSCALE=(SHIGH-SIDES-SIDES)/(Y1B-Y0B)
       SCALE=MIN(XSCALE,YSCALE)
       NCHAR=((X1B-X0B)*SCALE)/(HEIGHT*RATIO)-18
       IF (NCHAR.GE.1) THEN
            DO 36 I=1,79
                 IF (I.LE.NCHAR) THEN
                      TITLE(I:I)='_'
                 ELSE
                      TITLE(I:I)='X'
                 ENDIF
   36       CONTINUE            
            WRITE(6,37)NCHAR,TITLE
   37       FORMAT(/' Enter title for plot',
     +             ' (up to ',I2,' characters):'
     +              /' ',A79)
            WRITE (6,38)
   38       FORMAT ('+'\)
            TITLE='                                        '//
     +            '                                       '
            READ (5,39) TITLE
   39       FORMAT (A79)
       ENDIF
       WRITE (*,*)
       XOFSET=SWIDE/2.-SCALE*(X0B+X1B)/2.
       YOFSET=SHIGH/2.-SCALE*(Y0B+Y1B)/2.
       FILE=LFILE
       CALL PLOT(X0B,Y0B,UP)
       CALL PLOT(X1B,Y0B,DOWN)
       CALL PLOT(X1B,Y1B,DOWN)
       CALL PLOT(X0B,Y1B,DOWN)
       CALL PLOT(X0B,Y0B,DOWN)
       FILE=PFILE
C
C      CALL SYMBOL(X0B,Y0B,HEIGHT,INTEG,ANGLE,ICODE)
C      This call used to mark the lower-left corner (before text was added)
C
       XT=X0B*SCALE+XOFSET
       YT=Y1B*SCALE+YOFSET+(SIDES-HEIGHT)/2.
       WRITE (TFILE,40)XT,YT,HEIGHT,Y1B
   40  FORMAT('T ',2F10.4,'       0.0  ',F10.4,'       10'/1P,E10.3)
       YT=Y0B*SCALE+YOFSET-HEIGHT-(SIDES-HEIGHT)/2.
       IF (Y0B.EQ.0.0) THEN
            WRITE(TFILE,41)XT,YT,HEIGHT
   41       FORMAT('T ',2F10.4,'       0.0 ',F10.4,'       8'
     +             /'     0.0')
       ELSE
            WRITE (TFILE,40)XT,YT,HEIGHT,Y0B
       ENDIF
       XT=X0B*SCALE+XOFSET-HEIGHT-(SIDES-HEIGHT)/2.
       YT=SHIGH/2.+6.*HEIGHT*RATIO
       IF (X0B.EQ.0.0) THEN
            WRITE(TFILE,42)XT,YT,HEIGHT
   42       FORMAT('T ',2F10.4,'     -90.0 ',F10.4,'        8'
     +             /'     0.0')
       ELSE
            WRITE (TFILE,43)XT,YT,HEIGHT,X0B
   43       FORMAT('T ',2F10.4,'     -90.0 ',F10.4,'     10'/1P,E10.3)
       ENDIF
       XT=X1B*SCALE+XOFSET+(SIDES-HEIGHT)/2.
       WRITE (TFILE,43)XT,YT,HEIGHT,X1B
       XT=X0B*SCALE+XOFSET+14.*HEIGHT*RATIO
       YT=Y0B*SCALE+YOFSET-HEIGHT-(SIDES-HEIGHT)/2.
       WRITE(TFILE,44)XT,YT,HEIGHT,DFCON
   44  FORMAT('T ',2F10.4,'     0.0 ',F10.4,'     21'
     +       /'Interval = ',1P,E10.3)
       IF (NCHAR.GE.1) THEN
            YT=Y1B*SCALE+YOFSET+(SIDES-HEIGHT)/2.
            WRITE(TFILE,45)XT,YT,HEIGHT,NCHAR,TRIM(TITLE)
   45       FORMAT('T ',2F10.4,'     0.0 ',F10.4,I10
     +            /A)
       ENDIF
       CONLOW=HUGE
       CONHI= -HUGE
C
C   CONTOUR SELECTED REGION,
C   WORKING BY QUADRANGLES OF SIZE 2*DX BY 2*DY
C   EACH QUADRANGLE IS MADE UP OF TWO TRIANGULAR ELEMENTS
C
       NUMEL=(NCX2-NCX1)*(NCY2-NCY1)/2
       IDONE=0
       WRITE(6,60)
   60  FORMAT(' ')
       DO 100 IYLL=NCY2-2,NCY1,-2
            DO 90 JXLL=NCX1,NCX2-2,2
                 X1=XMIN+(JXLL-1)*DX
                 X2=X1+2.*DX
                 X3=X1
                 X4=X1+DX
                 X5=X4
                 X6=X1
                 Y1=YMIN+(IYLL-1)*DY
                 Y2=Y1+2.*DY
                 Y3=Y2
                 Y4=Y1+DY
                 Y5=Y2
                 Y6=Y4
                 F1=DATA(IYLL,  JXLL  )
                 F2=DATA(IYLL+2,JXLL+2)
                 F3=DATA(IYLL+2,JXLL  )
                 F4=DATA(IYLL+1,JXLL+1)
                 F5=DATA(IYLL+2,JXLL+1)
                 F6=DATA(IYLL+1,JXLL  )
                 CALL  CONTEL (INPUT,F1,F2,F3,F4,F5,F6,
     &                               X1,X2,X3,X4,X5,X6,
     &                               Y1,Y2,Y3,Y4,Y5,Y6,
     &                               DFCON,TICSIZ,
     &                        MODIFY,CONLOW,CONHI,LFILE,
     &                               XBOT,YBOT,ZBOT,XTOP,YTOP,ZTOP)
                 IDONE=IDONE+1
                 NEMI=NUMEL-IDONE
                 WRITE (*,89) IDONE, NEMI
C
                 X1=XMIN+(JXLL+1)*DX
                 X2=X1
                 X3=X1-2.*DX
                 X4=X1
                 X5=X1-DX
                 X6=X5
                 Y1=YMIN+(IYLL-1)*DY
                 Y2=Y1+2.*DY
                 Y3=Y1
                 Y4=Y1+DY
                 Y5=Y4
                 Y6=Y1
                 F1=DATA(IYLL  ,JXLL+2)
                 F2=DATA(IYLL+2,JXLL+2)
                 F3=DATA(IYLL  ,JXLL  )
                 F4=DATA(IYLL+1,JXLL+2)
                 F5=DATA(IYLL+1,JXLL+1)
                 F6=DATA(IYLL  ,JXLL+1)
                 CALL  CONTEL (INPUT,F1,F2,F3,F4,F5,F6,
     &                               X1,X2,X3,X4,X5,X6,
     &                               Y1,Y2,Y3,Y4,Y5,Y6,
     &                               DFCON,TICSIZ,
     &                        MODIFY,CONLOW,CONHI,LFILE,
     &                               XBOT,YBOT,ZBOT,XTOP,YTOP,ZTOP)
                 IDONE=IDONE+1
                 NEMI=NUMEL-IDONE
                 WRITE(6,89)IDONE,NEMI
   89            FORMAT('+',I6,' Elements contoured; ',I6,' to go....')
   90       CONTINUE
  100  CONTINUE
C
C   PLOT EXTREME VALUES FOUND
C
       Y0BI=Y0B*SCALE+YOFSET
       Y1BI=Y1B*SCALE+YOFSET
       XBOT=XBOT*SCALE+XOFSET
       YBOT=YBOT*SCALE+YOFSET
       WRITE(PFILE,110)XBOT,YBOT
  110  FORMAT('P ',2F8.3)
       IF (ABS(YBOT-Y1BI).LT.0.15) THEN
            YT=YBOT-0.25
       ELSE IF (ABS(YBOT-Y0BI).LT.0.15) THEN
            YT=YBOT+0.05
       ELSE
            YT=YBOT-0.10
       ENDIF
       IF (XBOT.LT.SWIDE/2.) THEN
            XT=XBOT
       ELSE
            XT=XBOT-10*HEIGHT*0.7
       ENDIF
       WRITE (TFILE,40)XT,YT,HEIGHT,ZBOT
       XTOP=XTOP*SCALE+XOFSET
       YTOP=YTOP*SCALE+YOFSET
       WRITE(PFILE,110)XTOP,YTOP
       IF (ABS(YTOP-Y1BI).LT.0.15) THEN
            YT=YTOP-0.25
       ELSE IF (ABS(YTOP-Y0BI).LT.0.15) THEN
            YT=YTOP+0.05
       ELSE
            YT=YTOP-0.10
       ENDIF
       IF (XTOP.LT.SWIDE/2.) THEN
            XT=XTOP
       ELSE
            XT=XTOP-10*HEIGHT*0.7
       ENDIF
       WRITE (TFILE,40)XT,YT,HEIGHT,ZTOP
C       
       CLOSE(UNIT=LFILE)
       WRITE(6,150)CONLOW,CONHI
  150  FORMAT(/' Contours on plot range from ',1P,E13.5,' to',E13.5
     &       //' (Run MakePlot to see results!)'/)
       CALL PAUSE()
       STOP
       END
C
C
C
       SUBROUTINE CONTEL (INPUT,F1,F2,F3,F4,F5,F6,
     &                          X1,X2,X3,X4,X5,X6,
     &                          Y1,Y2,Y3,Y4,Y5,Y6,
     &                          DFCON0,TICSIZ,
     &                   MODIFY,CONLOW,CONHI,LFILE,
     &                          XBOT,YBOT,ZBOT,XTOP,YTOP,ZTOP)
C
C   CONTOURS A SCALAR FIELD IN ONE ELEMENT OF A FINITE ELEMENT GRID.
C
C   THE 6 NODES OF THE TRIANGLE (3 VERTICES AND 3 MIDPOINTS OF SIDES)
C   ARE GENERICALLY NUMBERED 1..6, WITH 1..3 REFERRING TO THE 3 VERTICES IN
C   COUNTERCLOCKWISE ORDER FROM AN ARBITRARY STARTING POINT.
C   THEN NODE 4 IS OPPOSITE VERTEX 3,
C        NODE 5 IS OPPOSITE VERTEX 1,
C   AND  NODE 6 IS OPPOSITE VERTEX 2.
C   F1..F6 ARE THE VALUES OF THE SCALAR TO BE CONTOURED, AT THESE NODES.
C   X1..X6 ARE THE X-COORDINATES OF THESE NODES.
C   Y1..Y6 ARE THE Y-COORDINATES OF THESE NODES.
C   CONLOW AND CONHI ARE THE EXTREME VALUES FOUND IN CONTOURING.
C   DFCON IS THE DESIRED CONTOUR INTERVAL.
C   LFILE IS THE OUTPUT DEVICE UNIT FOR THE GRAPHICS METAFILE.
C   TICSIZ IS THE DESIRED LENGTH OF DOWN-SLOP TIC MARKS, IN
C      THE SAME UNITS AS THE MAP COORDINATES X1...X6 AND Y1...Y6.
C
       PARAMETER (NSTART=500)
       INTEGER DOWN,FILE,LFILE,UP
       LOGICAL CENTER,DASHED,HITLIM,GONOUT,CIRCLE,FINISH,DONE,UPLEFT
       DIMENSION PS(4,NSTART),DS(3),FN(6),DONE(NSTART),
     +           XLIST(6),YLIST(6)
       COMMON /IODEV/ FILE
       COMMON /SIZE/ SCALE,XOFSET,YOFSET
       DATA UP/3/, DOWN/2/
       DATA DSTEP/0.075/, NSPTIC/8/
       DATA I/0/
C
C   FUNCTION STATEMENT GIVING THE INTERPOLATION POLYNOMIALS
C   IN TERMS OF THE INTERNAL TRIANGULAR COORDINATES
C   (TERNARY DIAGRAM COORDINATES: S1+S2+S3=1).
C
       PHIVAL(S1,S2,S3,F1,F2,F3,F4,F5,F6)=
     +        F1*(-S1+2.*S1**2)+
     +        F2*(-S2+2.*S2**2)+
     +        F3*(-S3+2.*S3**2)+
     +        F4*(4.*S1*S2)+
     +        F5*(4.*S2*S3)+
     +        F6*(4.*S3*S1)
       FILE=LFILE
       LIMINT=4./DSTEP
       CENTER=.FALSE.
       I=I+1
       FTOPS=MAX(F1,F2,F3,F4,F5,F6)
       FLOOR=MIN(F1,F2,F3,F4,F5,F6)
       DFCON=DFCON0
   10  CNUM=(FTOPS-FLOOR)/DFCON
C
C SKIP ELEMENT IF MAXIMUM RANGE IS LESS THAN 0.01 CONTOUR INTERVAL
C
       IF(CNUM.LT.0.01) RETURN
C
C REDUCE CONTOUR INTERVAL IF THERE WOULD BE OVER 30 CONTOURS IN ELEMENT
C
       IF (CNUM.GT.32.) THEN
            DFCON=DFCON*10.
            GO TO 10
       ENDIF
C
       XLIST(1)=X1
       XLIST(2)=X2
       XLIST(3)=X3
       XLIST(4)=X4
       XLIST(5)=X5
       XLIST(6)=X6
       YLIST(1)=Y1
       YLIST(2)=Y2
       YLIST(3)=Y3
       YLIST(4)=Y4
       YLIST(5)=Y5
       YLIST(6)=Y6
       FN(1)=F1
       FN(2)=F2
       FN(3)=F3
       FN(4)=F4
       FN(5)=F5
       FN(6)=F6
C
C EXAMINE SIDES FOR EXTREMA AND MARK CONTOUR INTERSECTIONS
C
       NPS=0
       DO 100 MSIDE=1,3
            N1=MSIDE
            N2=MOD(MSIDE,3)+1
            NM=MSIDE+3
            FMAX=MAX(FN(N1),FN(N2))
            FMIN=MIN(FN(N1),FN(N2))
            IF (FMIN.LT.ZBOT) THEN
                 ZBOT=FMIN
                 NLOW=N1
                 IF (FN(N2).LT.FN(N1)) NLOW=N2
                 XBOT=XLIST(NLOW)
                 YBOT=YLIST(NLOW)
            ENDIF
            IF (FMAX.GT.ZTOP) THEN
                 ZTOP=FMAX
                 NHI=N1
                 IF (FN(N2).GT.FN(N1)) NHI=N2
                 XTOP=XLIST(NHI)
                 YTOP=YLIST(NHI)
            ENDIF
            DFDS1= -3.*FN(N1)+4.*FN(NM)-   FN(N2)
            DFDS2=     FN(N1)-4.*FN(NM)+3.*FN(N2)
            IF (DFDS1*DFDS2.GE.0.) THEN
                 CALL DOSIDE (FMAX,FMIN,DFCON,FN,N1,N2,NM,PS,NPS)
            ELSE
                 D2FDS=4.*FN(N1)-8.*FN(NM)+4.*FN(N2)
                 SEXT= -DFDS1/D2FDS
                 DS(N2)=SEXT
                 DS(N1)=1.00-SEXT
                 DS(6-N1-N2)=0.
                 FEXT=FN(N1)+
     &                DFDS1*SEXT+
     &                0.5*D2FDS*SEXT**2
                 XEXT=PHIVAL(DS(1),DS(2),DS(3),X1,X2,X3,X4,X5,X6)
                 YEXT=PHIVAL(DS(1),DS(2),DS(3),Y1,Y2,Y3,Y4,Y5,Y6)
C
C FIND INTERSECTIONS OF CONTOURS WITH SIDE CONTAINING EXTREMUM
C
                 FMAX=MAX(FN(N1),FEXT)
                 FMIN=MIN(FN(N1),FEXT)
                 CALL DOPART (FEXT,FMAX,FMIN,DFCON,FN,
     &                        N1,N2,NM,0.,SEXT,PS,NPS)
                 FMAX=MAX(FEXT,FN(N2))
                 FMIN=MIN(FEXT,FN(N2))
                 CALL DOPART (FEXT,FMAX,FMIN,DFCON,FN,
     &                        N1,N2,NM,SEXT,1.,PS,NPS)
                 IF (FEXT.LT.ZBOT) THEN
                      ZBOT=FEXT
                      XBOT=XEXT
                      YBOT=YEXT
                 ENDIF
                 IF (FEXT.GT.ZTOP) THEN
                      ZTOP=FEXT
                      XTOP=XEXT
                      YTOP=YEXT
                 ENDIF
            ENDIF
  100  CONTINUE
C
C SEARCH FOR EXTREMUM WITHIN INTERIOR OF ELEMENT
C
       CDET=16.*(-F6**2+2.*F6*F5+2.*F6*F4-2.*F6*F2-F5**2+2.*F5*F4
     +         -2.*F5*F1-F4**2-2.*F4*F3+F3*F2+F3*F1+F2*F1)
       IF (ABS(CDET).LT. 1.E-25) GO TO 1000
       S2EXT=(4.*F6**2-4.*F6*F5-4.*F6*F4-F6*F3+2.*F6*F2-F6*F1+F5*
     +      F3+3.*F5*F1+3.*F4*F3+F4*F1-F3*F2-2.*F3*F1-F2*F1)/(4.*(F6
     +      **2-2.*F6*F5-2.*F6*F4+2.*F6*F2+F5**2-2.*F5*F4+2.*F5*F1+F4
     +      **2+2.*F4*F3-F3*F2-F3*F1-F2*F1))
       S3EXT=(-4.*F6*F4+3.*F6*F2+F6*F1-4.*F5*F4+F5*F2+3.*F5*F1+4.
     +      *F4**2+2.*F4*F3-F4*F2-F4*F1-F3*F2-F3*F1-2.*F2*F1)/(4.*(F6
     +      **2-2.*F6*F5-2.*F6*F4+2.*F6*F2+F5**2-2.*F5*F4+2.*F5*F1+F4
     +      **2+2.*F4*F3-F3*F2-F3*F1-F2*F1))
       S1EXT=1.0-S2EXT-S3EXT
       IF (S1EXT.GT.0.99999.OR.S1EXT.LT.0.00001) GO TO 1000
       IF (S2EXT.GT.0.99999.OR.S2EXT.LT.0.00001) GO TO 1000
       IF (S3EXT.GT.0.99999.OR.S3EXT.LT.0.00001) GO TO 1000
       XEXT=PHIVAL(S1EXT,S2EXT,S3EXT,X1,X2,X3,X4,X5,X6)
       YEXT=PHIVAL(S1EXT,S2EXT,S3EXT,Y1,Y2,Y3,Y4,Y5,Y6)
       FEXT=PHIVAL(S1EXT,S2EXT,S3EXT,F1,F2,F3,F4,F5,F6)
       IF (FEXT.LT.ZBOT) THEN
            ZBOT=FEXT
            XBOT=XEXT
            YBOT=YEXT
       ENDIF
       IF (FEXT.GT.ZTOP) THEN
            ZTOP=FEXT
            XTOP=XEXT
            YTOP=YEXT
       ENDIF
C
C DON'T COUNT SADDLE POINTS
C
       DISCA=F1-2.*F4+F2
       DISCB=F2-2.*F5+F3
       DISCC=F3-2.*F6+F1
       CENTER=((DISCA.GT.0.).AND.(DISCB.GT.0.).AND.(DISCC.GT.0.))
     +    .OR.((DISCA.LT.0.).AND.(DISCB.LT.0.).AND.(DISCC.LT.0.))
       IF (.NOT.CENTER) GO TO 1000
C
C FIND CONTOUR STARTING/STOPPING POINT ALONG CHORD FROM A NODE TO EXT.
C
       NCL=1
       DIFF=ABS(F1-FEXT)
       DS(1)=S1EXT-1.
       DS(2)=S2EXT
       DS(3)=S3EXT
       DO 120 J=2,6
            DFF=ABS(FN(J)-FEXT)
            IF (DFF.LT.DIFF) THEN
                 NCL=J
                 DIFF=DFF
                 DS(1)=S1EXT
                 DS(2)=S2EXT
                 DS(3)=S3EXT
                 IF (J.EQ.2) DS(2)=S2EXT-1.
                 IF (J.EQ.3) DS(3)=S3EXT-1.
                 IF (J.EQ.4.OR.J.EQ.6) DS(1)=S1EXT-0.5
                 IF (J.EQ.4.OR.J.EQ.5) DS(2)=S2EXT-0.5
                 IF (J.EQ.5.OR.J.EQ.6) DS(3)=S3EXT-0.5
            ENDIF
  120  CONTINUE
       CALL DOLINE (FEXT,DFCON,FN,NCL,DS,S1EXT,S2EXT,S3EXT,
     +              PS,NPS)
C
C   RESUME HERE IF THERE IS NO INTERIOR EXTREMUM
C
 1000  IF (NPS.EQ.0) RETURN
C
C   KEEP TRACK OF WHETHER PARTICULAR CONTOUR ENDPOINTS HAVE BEEN USED
C
       DO 1150 K=1,NPS
            DONE(K)=.FALSE.
 1150  CONTINUE
       DO 9000 N=1,NPS
C
C ACTUALLY PLOT CONTOUR FROM ONE PARTICULAR STARTING-POINT
C
            IF (.NOT.DONE(N)) THEN
C
C   INITIALIZE INTEGRATION OF CONTOUR
C
                 DONE(N)=.TRUE.
                 S1=PS(1,N)
                 S2=PS(2,N)
                 S3=PS(3,N)
                 S1OLD=S1
                 S2OLD=S2
                 S3OLD=S3
                 X=PHIVAL(S1,S2,S3,X1,X2,X3,X4,X5,X6)
                 Y=PHIVAL(S1,S2,S3,Y1,Y2,Y3,Y4,Y5,Y6)
C
C   PLACE PEN AT INITIAL POINT
C
                 CALL PLOT(X,Y,UP)
                 FVALUE=PS(4,N)
                 DASHED=(FVALUE.EQ.0.)
                 LASTPN=UP
                 CONLOW=AMIN1(CONLOW,FVALUE)
                 CONHI=AMAX1(CONHI,FVALUE)
                 ANGLE=0.
                 IF (CENTER) ANGLE=ATAN2((Y-YEXT),(X-XEXT))
                 ANGLEP=ANGLE
                 ROT=0.
                 DFDS2=-4.*S3*F6+4.*S3*F5-4.*S3*F4+
     &                  4.*S3*F1-8.*S2*F4+4.*S2*F2+
     &                  4.*S2*F1+4.*F4-F2-3.*F1
                 DFDS3=-8.*S3*F6+4.*S3*F3+4.*S3*F1
     &                 -4.*S2*F6+4.*S2*F5-4.*S2*
     &                  F4+4.*S2*F1+4.*F6-F3-3.*F1
                 GRADF=SQRT(DFDS2**2+DFDS3**2)
                 IF (GRADF.EQ.0.) GO TO 8999
                 GRADFX=DFDS2/GRADF
                 GRADFY=DFDS3/GRADF
                 ROUNDX= +GRADFY
                 ROUNDY= -GRADFX
                 DS2=ROUNDX*DSTEP*0.1
                 DS3=ROUNDY*DSTEP*0.1
                 S2P=S2+DS2
                 S3P=S3+DS3
                 S1P=1.00-S2P-S3P
C
C REVERSE INTEGRATION STEP DIRECTION IF CONTOUR POINTS OUTWARD
C
                 IF ((S1P.LT.0..OR.S1P.GT.1.)
     +           .OR.(S2P.LT.0..OR.S2P.GT.1.)
     +           .OR.(S3P.LT.0..OR.S3P.GT.1.)) THEN
                      COUNTR= -1.
                      UPLEFT=.FALSE.
                 ELSE
                      UPLEFT=.TRUE.
                      COUNTR=1.
                 ENDIF
                 NSEG=0
C
C  BEGIN OR CONTINUE LOOP OF INTEGRATION OF CONTOUR LINE
C
                 ITIC=0
 3000            NSEG=NSEG+1
                 ITIC=ITIC+1
C
C  PLOT A DOWNSLOPE TIC MARK FROM TIME TO TIME
C
                 IF (MOD(ITIC,NSPTIC).EQ.(NSPTIC-1)) THEN
                      FORWAR=ATAN2((Y-YLAST),(X-XLAST))
                      IF (UPLEFT) THEN
                           TICANG=FORWAR-1.5708
                      ELSE
                           TICANG=FORWAR+1.5708
                      ENDIF
                      XT=X+TICSIZ*COS(TICANG)
                      YT=Y+TICSIZ*SIN(TICANG)
                      CALL PLOT(XT,YT,DOWN)
C                  (KEEP CONTOUR CONTINUOUS FOR GOOD METAFILE STRUCTURE)
                      CALL PLOT(X ,Y ,DOWN)
                 ENDIF
C
C  EXTRAPOLATE TO NEXT POINT BY EXPLICIT METHOD
C
                 DS2=ROUNDX*COUNTR*DSTEP
                 DS3=ROUNDY*COUNTR*DSTEP
                 S2P=S2+DS2
                 S3P=S3+DS3
                 S1P=1.00-S2P-S3P
C
C  RECOMPUTE SAME STEP BY BACKWARD METHOD
C
                 DFDS2=-4.*S3P*F6+4.*S3P*F5-4.*S3P*F4
     +                 +4.*S3P*F1-8.*S2P*F4+4.*S2P*
     +                  F2+4.*S2P*F1+4.*F4-F2-3.*F1
                 DFDS3=-8.*S3P*F6+4.*S3P*F3+4.*S3P*F1
     +                 -4.*S2P*F6+4.*S2P*F5-4.*S2P*
     +                   F4+4.*S2P*F1+4.*F6-F3-3.*F1
                 GRADF=SQRT(DFDS2**2+DFDS3**2)
                 IF (GRADF.EQ.0.) GO TO 8999
                 GRADFX=DFDS2/GRADF
                 GRADFY=DFDS3/GRADF
                 ROUNDX= +GRADFY
                 ROUNDY= -GRADFX
                 DS2P=ROUNDX*DSTEP*COUNTR
                 DS3P=ROUNDY*DSTEP*COUNTR
C
C  ACTUAL INTEGRATION STEP BY TRAPEZOIDAL METHOD
C
                 DS2=0.5*(DS2+DS2P)
                 DS3=0.5*(DS3+DS3P)
                 DSLEN=SQRT(DS2**2+DS3**2)
                 IF ((DSLEN/DSTEP).LT.0.10) GO TO 8999
                 S2=S2+DS2
                 S3=S3+DS3
                 S1=1.00-S2-S3
C
C  CORRECT CONTOUR TO ACTUAL F VALUE DESIRED
C
                 TRIAL=PHIVAL(S1,S2,S3,F1,F2,F3,F4,F5,F6)
                 ERROR=TRIAL-FVALUE
                 IF (ABS(ERROR).GE.DFCON) GO TO 8999
                 DFDS2=-4.*S3 *F6+4.*S3 *F5-4.*S3 *F4
     +                 +4.*S3 *F1-8.*S2 *F4+4.*S2 *
     +                  F2+4.*S2 *F1+4.*F4-F2-3.*F1
                 DFDS3=-8.*S3 *F6+4.*S3 *F3+4.*S3 *F1
     +                 -4.*S2 *F6+4.*S2 *F5-4.*S2 *
     +                  F4+4.*S2 *F1+4.*F6-F3-3.*F1
                 GRADF=SQRT(DFDS2**2+DFDS3**2)
                 IF (GRADF.EQ.0.) GO TO 8999
                 GRADFX=DFDS2/GRADF
                 GRADFY=DFDS3/GRADF
                 DISTNC= -ERROR/GRADF
                 IF (ABS(DISTNC).GT.DSTEP) DISTNC=
     +                   DISTNC*DSTEP/ABS(DISTNC)
                 S2=S2+DISTNC*GRADFX
                 S3=S3+DISTNC*GRADFY
                 S1=1.00-S2-S3
C
C   DECIDE WHETHER CONTOUR IS FINISHED OR NOT
C
                 HITLIM=NSEG.GE.LIMINT
                 IF (HITLIM) WRITE(6,3501)FVALUE,I
 3501            FORMAT(' ',1PE10.2,' CONTOUR IN ELEMENT ',I3,
     +                    ' SEEMS TO BE IN LOOP.     TERMINATED.')
                 GONOUT=(S1.LT.0..OR.S1.GT.1.).OR.
     +                  (S2.LT.0..OR.S2.GT.1.).OR.
     +                  (S3.LT.0..OR.S3.GT.1.)
                 FINISH=GONOUT.OR.HITLIM
                 IF (CENTER) THEN
                      XT=PHIVAL(S1,S2,S3,X1,X2,X3,X4,X5,X6)
                      YT=PHIVAL(S1,S2,S3,Y1,Y2,Y3,Y4,Y5,Y6)
                      ANGLEP=ATAN2((YT-YEXT),(XT-XEXT))
                      DROT=MIN(ABS(ANGLEP-ANGLE),
     +                  6.2832-ABS(ANGLEP-ANGLE))
                      ROT=ROT+DROT
                      CIRCLE=ROT.GE.6.2832
                      FINISH=FINISH.OR.CIRCLE
                 ENDIF
C
C  SHORTEN VECTOR IF IT EXTENDS OUTSIDE OF THE ELEMENT
C
                 IF (GONOUT) THEN
                      RAT=1.0
                      IF(S1.GT.1.)RAT=
     &                    MIN(RAT,((1.-S1OLD)/(S1-S1OLD)))
                      IF(S2.GT.1.)RAT=
     &                    MIN(RAT,((1.-S2OLD)/(S2-S2OLD)))
                      IF(S3.GT.1.)RAT=
     &                    MIN(RAT,((1.-S3OLD)/(S3-S3OLD)))
                      IF(S1.LT.0.)RAT=
     &                    MIN(RAT,((0.-S1OLD)/(S1-S1OLD)))
                      IF(S2.LT.0.)RAT=
     &                    MIN(RAT,((0.-S2OLD)/(S2-S2OLD)))
                      IF(S3.LT.0.)RAT=
     &                    MIN(RAT,((0.-S3OLD)/(S3-S3OLD)))
                      RAT=MAX(RAT,0.0)
                      S2=S2OLD+(S2-S2OLD)*RAT
                      S3=S3OLD+(S3-S3OLD)*RAT
                      S1=1.00-S2-S3
                 ENDIF
C
C CHECK WHETHER ANY OTHER POINTS HAVE BEEN CONTACTED
C
                 IF (N.LT.NPS) THEN
                      NP1=N+1
                      DO 4000 M=NP1,NPS
                           IF ((.NOT.DONE(M)).AND.(PS(4,M).EQ.FVALUE)
     +                          .AND.(ABS(PS(1,M)-S1).LE.DSTEP)
     +                          .AND.(ABS(PS(2,M)-S2).LE.DSTEP)
     +                          .AND.(ABS(PS(3,M)-S3).LE.DSTEP)) THEN
                                 DONE(M)=.TRUE.
                                 GO TO 4001
                           ENDIF
 4000                 CONTINUE
 4001                 CONTINUE
                 ENDIF
C
C   SAVE PREVIOUS POINT
C
                 XLAST=X
                 YLAST=Y
C
C   LOCATE (X,Y) COORDINATES OF NEXT POINT
C
                 X=PHIVAL(S1,S2,S3,X1,X2,X3,X4,X5,X6)
                 Y=PHIVAL(S1,S2,S3,Y1,Y2,Y3,Y4,Y5,Y6)
C
C   CHECK WHETHER PEN SHOULD BE LIFTED TO DASH 0 CONTOURS
C
                 IF (DASHED) THEN
                      IF (LASTPN.EQ.UP) THEN
                           IPEN=DOWN
                      ELSE
                           IPEN=UP
                      ENDIF
                 ELSE
                      IPEN=DOWN
                 ENDIF
                 LASTPN=IPEN
C
C************************************************
C  ACTUALLY PLOT VECTOR
C
                 CALL PLOT(X,Y,IPEN)
C
C************************************************
C
C  PREPARE FOR NEXT ITERATION IF ONE IS NEEDED
C
                 S1OLD=S1
                 S2OLD=S2
                 S3OLD=S3
                 IF (.NOT.FINISH) THEN
                      DFDS2=-4.*S3 *F6+4.*S3 *F5-4.*S3 *F4
     +                      +4.*S3 *F1-8.*S2 *F4+4.*S2 *
     +                       F2+4.*S2 *F1+4.*F4-F2-3.*F1
                      DFDS3=-8.*S3 *F6+4.*S3 *F3+4.*S3 *F1
     +                      -4.*S2 *F6+4.*S2 *F5-4.*S2 *
     +                       F4+4.*S2 *F1+4.*F6-F3-3.*F1
                      GRADF=SQRT(DFDS2**2+DFDS3**2)
                      IF (GRADF.EQ.0.) GO TO 8999
                      GRADFX=DFDS2/GRADF
                      GRADFY=DFDS3/GRADF
                      ROUNDX= +GRADFY
                      ROUNDY= -GRADFX
                      ANGLE=ANGLEP
                      GO TO 3000
                 ENDIF
 8999            CONTINUE
            ENDIF
 9000  CONTINUE
       RETURN
       END
C
C
C
      SUBROUTINE PLOT (X,Y,IPEN)
C
C   SUBSTITUTE FOR THE CALCOMP/PLOT88 ROUTINE OF SAME NAME.
C   DRAWS A LINE SEGMENT FROM PREVIOUS PEN POSITION TO
C   SPECIFIED PEN POSITION IF IPEN INDICATES PEN IS DOWN.
C
       INTEGER DOWN,FILE,UP
       COMMON /IODEV/ FILE
       COMMON /OLDXY/ OLDX,OLDY
       COMMON /SIZE/ SCALE,XOFSET,YOFSET
       DATA UP/3/, DOWN/2/
       XS=X*SCALE+XOFSET
       YS=Y*SCALE+YOFSET
       IF (IPEN.EQ.DOWN) THEN
            WRITE(FILE,1)OLDX,OLDY,XS,YS
    1       FORMAT('L ',4F8.3)
       ENDIF
       OLDX=XS
       OLDY=YS
       RETURN
       END
C
C
C
      SUBROUTINE SYMBOL (X,Y,HEIGHT,INTEG,ANGLE,ICODE)
C
C   SUBSTITUTE FOR THE CALCOMP/PLOT88 ROUTINE OF SAME NAME.
C   WRITES A LINE TO THE POINTS METAFILE, EVENTUALLY CAUSING
C   A CENTERED + SYMBOL AT COORDINATES (X,Y).
C   OTHER PARAMETERS ARE DUMMIES WITH NO FUNCTION...YET.
C
       INTEGER FILE
       COMMON /IODEV/ FILE
       COMMON /SIZE/ SCALE,XOFSET,YOFSET
       XS=X*SCALE+XOFSET
       YS=Y*SCALE+YOFSET
       WRITE(FILE,1)XS,YS
    1  FORMAT('P ',2F8.3)
       RETURN
       END
C
C
C
       SUBROUTINE DOSIDE (FMAX,FMIN,DFCON,FN,N1,N2,NM,PS,NPS)
C
C   FIND BEGINNING/END POINTS OF CONTOURS ALONG SIDE OF ELEMENT
C
       DIMENSION FN(6),PS(4,500)
       ILOW=IBELOW(FMIN/DFCON)
       IF (FMIN.EQ.(DFCON*ILOW)) ILOW=ILOW-1
       IHI=IABOVE(FMAX/DFCON)
       IF (FMAX.EQ.(DFCON*IHI)) IHI=IHI+1
       NBTWEN=IHI-ILOW-1
       IF (NBTWEN.GE.1) THEN
            NBASE=ILOW
            A=2.*FN(N1)-4.*FN(NM)+2.*FN(N2)
            B= -3.*FN(N1)+4.*FN(NM)-FN(N2)
            DO 10 K=1,NBTWEN
                 N=K+NBASE
                 F=N*DFCON
                 C= FN(N1)-F
                 IF (A.NE.0.) THEN
                      DISC=SQRT(MAX(B**2-4.*A*C,0.0))
                      ROOT1=(-B+DISC)/(2.*A)
                      ROOT2=(-B-DISC)/(2.*A)
                      OUT1=MAX(MAX(0.,-ROOT1),MAX(ROOT1-1.,0.))
                      OUT2=MAX(MAX(0.,-ROOT2),MAX(ROOT2-1.,0.))
                      IF (OUT1.LE.OUT2) THEN
                           IF (OUT1.GT.0.1)GO TO 10
                           S=ROOT1
                      ELSE
                           IF (OUT2.GT.0.1)GO TO 10
                           S=ROOT2
                      ENDIF
                 ELSE IF (FN(N2).NE.FN(N1)) THEN
                      S=(F-FN(N1))/(FN(N2)-FN(N1))
                 ELSE
                      S=0.
                 ENDIF
                 NPS=NPS+1
                 PS(N1,NPS)=1.-S
                 PS(N2,NPS)=S
                 N3=6-N1-N2
                 PS(N3,NPS)=0.
                 PS(4,NPS)=F
   10       CONTINUE
       ENDIF
       RETURN
       END
C
C
C
       SUBROUTINE DOPART (FEXT,FMAX,FMIN,DFCON,FN,
     +                 N1,N2,NM,S1,S2,PS,NPS)
C
C   FIND CONTOUR END POINTS ALONG PART OF AN ELEMENT SIDE
C
       DIMENSION FN(6),PS(4,500)
       ILOW=IBELOW(FMIN/DFCON)
       IF (FMIN.EQ.(DFCON*ILOW)) ILOW=ILOW-1
       IHI=IABOVE(FMAX/DFCON)
       IF (FMAX.EQ.(DFCON*IHI)) IHI=IHI+1
       NBTWEN=IHI-ILOW-1
       IF (NBTWEN.GE.1) THEN
            NBASE=ILOW
            A=2.*FN(N1)-4.*FN(NM)+2.*FN(N2)
            B= -3.*FN(N1)+4.*FN(NM)-FN(N2)
            DO 20 K=1,NBTWEN
                 N=K+NBASE
                 F=N*DFCON
                 C= FN(N1)-F
                 IF (A.NE.0.) THEN
                       DISC=SQRT(MAX(B**2-4.*A*C,0.0))
                       ROOT1=(-B+DISC)/(2.*A)
                       ROOT2=(-B-DISC)/(2.*A)
                       OUT1=MAX(MAX(0.,S1-ROOT1),MAX(ROOT1-S2,0.))
                       OUT2=MAX(MAX(0.,S1-ROOT2),MAX(ROOT2-S2,0.))
                       IF (OUT1.LE.OUT2) THEN
                            IF (OUT1.GT.0.1)GO TO 20
                            S=ROOT1
                       ELSE
                            IF (OUT2.GT.0.1)GO TO 20
                            S=ROOT2
                       ENDIF
                  ELSE IF (FN(N1).NE.FN(N2)) THEN
                       S=(F-FN(N1))/(FN(N2)-FN(N1))
                  ELSE
                       S=0.
                  ENDIF
                  NPS=NPS+1
                  PS(N1,NPS)=1.-S
                  PS(N2,NPS)=S
                  N3=6-N1-N2
                  PS(N3,NPS)=0.
                  PS(4,NPS)=F
   20        CONTINUE
        ENDIF
        RETURN
        END
C
C
C
       SUBROUTINE DOLINE (FEXT,DFCON,FN,NCL,DS,S1EXT,S2EXT,S3EXT,
     +              PS,NPS)
C
C   FINDS CONTOUR START/END POINT ALONG LINE FROM A NODE TO AN EXTREMUM
C
       DIMENSION FN(6),DS(3),PS(4,500)
       PHIVAL(S1,S2,S3,F1,F2,F3,F4,F5,F6)=
     +           F1*(-S1+2.*S1**2)+
     +           F2*(-S2+2.*S2**2)+
     +           F3*(-S3+2.*S3**2)+
     +           F4*(4.*S1*S2)+
     +           F5*(4.*S2*S3)+
     +           F6*(4.*S3*S1)
       FMAX=AMAX1(FN(NCL),FEXT)
       FMIN=AMIN1(FN(NCL),FEXT)
       ILOW=IBELOW(FMIN/DFCON)
       IF (FMIN.EQ.(DFCON*ILOW)) ILOW=ILOW-1
       IHI=IABOVE(FMAX/DFCON)
       IF (FMAX.EQ.(DFCON*IHI)) IHI=IHI+1
       NBTWEN=IHI-ILOW-1
       IF (NBTWEN.GE.1) THEN
            NBASE=ILOW
            FMID=PHIVAL((S1EXT-0.5*DS(1)),(S2EXT-0.5*DS(2)),
     +                  (S3EXT-0.5*DS(3)),FN(1),FN(2),FN(3),
     +                   FN(4),FN(5),FN(6))
            A=2.*FN(NCL)-4.*FMID+2.*FEXT
            B= -3.*FN(NCL)+4.*FMID-FEXT
            DO 20 K=1,NBTWEN
                 N=K+NBASE
                 F=N*DFCON
                 C= FN(NCL)-F
                 IF (A.NE.0.) THEN
                      DISC=SQRT(MAX(B**2-4.*A*C,0.0))
                      ROOT1=(-B+DISC)/(2.*A)
                      ROOT2=(-B-DISC)/(2.*A)
                      OUT1=MAX(MAX(0.,-ROOT1),MAX(ROOT1-1.,0.))
                      OUT2=MAX(MAX(0.,-ROOT2),MAX(ROOT2-1.,0.))
                      IF (OUT1.LE.OUT2) THEN
                           S=ROOT1
                      ELSE
                           S=ROOT2
                      ENDIF
                 ELSE IF (FEXT.NE.FN(NCL)) THEN
                      S=(F-FN(NCL))/(FEXT-FN(NCL))
                 ELSE
                      S=0.
                 ENDIF
                 NPS=NPS+1
                 PS(1,NPS)=S1EXT-(1.-S)*DS(1)
                 PS(2,NPS)=S2EXT-(1.-S)*DS(2)
                 PS(3,NPS)=S3EXT-(1.-S)*DS(3)
                 PS(4,NPS)=F
   20       CONTINUE
       ENDIF
       RETURN
       END
C
       INTEGER FUNCTION IBELOW (X)
C
C   RETURNS INTEGER .LE. X, UNLIKE INT FUNCTION
C
       IBELOW=INT(X)
       IF (X.LT.(1.*IBELOW)) IBELOW=IBELOW-1
       RETURN
       END
C
       INTEGER FUNCTION IABOVE (X)
C
C   RETURNS INTEGER .GE. X, UNLIKE INT FUNCTION
C
       IF (X.LE.0.) THEN
            IABOVE=INT(X)
       ELSE
            IABOVE=INT(X)
            IF (X.GT.IABOVE) IABOVE=IABOVE+1
       ENDIF
       RETURN
       END
C
C
C   
       SUBROUTINE PAUSE()
       WRITE(*,"(/' Press [Enter] to continue...'\)")
       READ(*,*)
       RETURN
       END

